@@ -71,7 +71,7 @@
 // is enabled.
 //
 
-use guff::{GaloisField};
+use guff::*;
 
 // Only one x86 implementation, included automatically
 #[cfg(any(target_arch = "x86", target_arch = "x86_64"))]
@@ -104,17 +104,17 @@ pub fn _monomorph() {
 	0,0,0, 0,0,0, 0,0,1,
     ];
     let mut transform =	// mut because of iterator
-	X86SimpleMatrix::<x86::X86u8x16Long0x11b>::new(9,9,true);
+	Matrix::new(9,9,true);
     transform.fill(&identity[..]);
     
     // 17 is coprime to 9
     let mut input =
-	X86SimpleMatrix::<x86::X86u8x16Long0x11b>::new(9,17,false);
+	Matrix::new(9,17,false);
     let vec : Vec<u8> = (1u8..=9 * 17).collect();
     input.fill(&vec[..]);
 
     let mut output =
-	X86SimpleMatrix::<x86::X86u8x16Long0x11b>::new(9,17,false);
+	Matrix::new(9,17,false);
 
     // works if output is stored in colwise format
     inner_fn(&mut transform, &mut input, &mut output);
@@ -137,41 +137,38 @@ pub mod arm_vmull;
 #[cfg(feature = "simulator")]
 pub mod simulator;
 
-/// Greatest Common Divisor for two integers
-pub fn gcd(mut a : usize, mut b : usize) -> usize {
-  let mut t;
-    loop {
-	if b == 0 { return a }
-	t = b;
-	b = a % b;
-	a = t;
-    }
+
+// I had wanted to have different matrix implementations that offered
+// different ways of implementing read_next(). I have moved away from
+// that goal by moving state out of matrices, though. It may be better
+// to offer different multiply functions.
+//
+// My most immediate goal now is providing Matrix and multiply support
+// for Arm. I don't want to just copy/paste code, but that might be
+// the best solution for now. To support that, I'll define an
+// arch-dependent Matrix type:
+
+#[cfg(any(target_arch = "x86", target_arch = "x86_64"))]
+pub mod types {
+    pub type NativeSimd = crate::x86::X86u8x16Long0x11b;
+    pub type Matrix = crate::x86::X86Matrix<NativeSimd>;
 }
 
-/// Greatest Common Divisor for three integers
-pub fn gcd3(a : usize, b : usize, c: usize) -> usize {
-    gcd(a, gcd(b,c))
+#[cfg(all(any(target_arch = "aarch64", target_arch = "arm"),
+	  feature = "arm_vmull"))]
+pub mod types {
+    pub type NativeSimd = crate::arm_vmull::VmullEngine8x8;
+    pub type Matrix = crate::arm_vmull::ArmMatrix::<NativeSimd>;
 }
 
-/// Greatest Common Divisor for four integers
-pub fn gcd4(a : usize, b : usize, c: usize, d : usize) -> usize {
-    gcd(gcd(a,b), gcd(c,d))
-}
+pub use types::*;
 
-/// Least Common Multiple for two integers
-pub fn lcm(a : usize, b : usize) -> usize {
-    (a / gcd(a,b)) * b
-}
+// (actually, copy/paste worked with only type changes, so I can work
+// on making a more generic matrix)
 
-/// Least Common Multiple for three integers
-pub fn lcm3(a : usize, b : usize, c: usize) -> usize {
-    lcm( lcm(a,b), c)
-}
-
-/// Least Common Multiple for four integers
-pub fn lcm4(a : usize, b : usize, c: usize, d : usize) -> usize {
-    lcm( lcm(a,b), lcm(c,d) )
-}
+/// GCD and LCM functions
+pub mod numbers;
+pub use numbers::*;
 
 /// SIMD support, based on `simulator` module
 ///
@@ -182,6 +179,8 @@ pub trait Simd {
     type V;			// vector type, eg [u8; 8]
     const SIMD_BYTES : usize;
 
+    fn zero_vector() -> Self;
+
     fn cross_product(a : Self, b : Self) -> Self;
     unsafe fn sum_across_n(m0 : Self, m1 : Self, n : usize, off : usize)
 			   -> (Self::E, Self);
@@ -190,6 +189,26 @@ pub trait Simd {
     // alternative to using num_traits.
     fn zero_element() -> Self::E;
     fn add_elements(a : Self::E, b : Self::E) -> Self::E;
+
+
+    // moved from SimdMatrix
+    unsafe fn read_next(mod_index : &mut usize,
+			array_index : &mut usize,
+			array     : &[Self::E],
+			size      : usize,
+			ra_size : &mut usize,
+			ra : &mut Self)
+	-> Self
+    where Self : Sized;
+
+    /// load from memory (useful for testing, benchmarking)
+    unsafe fn from_ptr(ptr: *const Self::E) -> Self
+	where Self : Sized;
+
+    /// Cross product of two slices; useful for testing, benchmarking
+    /// Uses fixed poly at the moment.
+    fn cross_product_slices(dest: &mut [Self::E],
+			    av : &[Self::E], bv : &[Self::E]);
 }
 
 // For the SimdMatrix trait, I'm not going to distinguish between
@@ -223,12 +242,18 @@ pub trait SimdMatrix<S : Simd> {
     // means that re-using the xform can/will result in the
     // read_next() state being wrong. It doesn't matter so much for
     // input matrix, since fill() should reset state to zero.
-    fn reset(&mut self);
+    // fn reset(&mut self);
 
-    /// Wrap-around read of matrix, returning a Simd vector type
-    unsafe fn read_next(&mut self) -> S;
-    /// Wrap-around diagonal write of (output) matrix
-    // fn write_next(&mut self, val : S::E);
+    // Wrap-around read of matrix, returning a Simd vector type
+    // 
+    // unsafe fn read_next(&mut self) -> S; // moved to Simd
+    
+    // Wrap-around diagonal write of (output) matrix
+    // fn write_next(&mut self, val : S::E); // moved to matrix mul
+
+    
+
+    
     fn indexed_write(&mut self, index : usize, elem : S::E);
     fn as_mut_slice(&mut self) -> &mut [S::E];
     fn as_slice(&self) -> &[S::E];
@@ -284,27 +309,60 @@ pub unsafe fn simd_warm_multiply<S : Simd + Copy>(
     let mut sum         = S::zero_element();
     let simd_width = S::SIMD_BYTES;
 
+    // Code for read_next() that was handled in SimdMatrix has now
+    // moved to Simd. We need to track those variables here.
+    let mut xform_mod_index = 0;
+    let mut xform_array_index = 0;
+    let     xform_array = xform.as_slice();
+    let     xform_size  = xform.size();
+    let mut xform_ra_size = 0;
+    let mut xform_ra = S::zero_vector();
+
+    let mut input_mod_index = 0;
+    let mut input_array_index = 0;
+    let     input_array = input.as_slice();
+    let     input_size  = input.size();
+    let mut input_ra_size = 0;
+    let mut input_ra = S::zero_vector();
+
     // we handle or and oc (was in matrix class)
     let mut or : usize = 0;
     let mut oc : usize = 0;
     let orows = output.rows();
     let ocols = output.cols();
 
-    // reset read_next state in case matrices reused for multiply
-    xform.reset();
-    input.reset();
-
     // read ahead two products
 
     let mut i0 : S;
     let mut x0 : S;
 
-    x0 = xform.read_next();
-    i0 = input.read_next();
+    x0 = S::read_next(&mut xform_mod_index,
+			 &mut xform_array_index,
+			 xform_array,
+			 xform_size,
+			 &mut xform_ra_size,
+			 &mut xform_ra);
+    i0 = S::read_next(&mut input_mod_index,
+			 &mut input_array_index,
+			 input_array,
+			 input_size,
+			 &mut input_ra_size,
+			 &mut input_ra);
+
     let mut m0 = S::cross_product(x0,i0);
 
-    x0  = xform.read_next();
-    i0  = input.read_next();
+    x0 = S::read_next(&mut xform_mod_index,
+			 &mut xform_array_index,
+			 xform_array,
+			 xform_size,
+			 &mut xform_ra_size,
+			 &mut xform_ra);
+    i0 = S::read_next(&mut input_mod_index,
+			 &mut input_array_index,
+			 input_array,
+			 input_size,
+			 &mut input_ra_size,
+			 &mut input_ra);
     let mut m1  = S::cross_product(x0,i0);
 
     let mut offset_mod_simd = 0;
@@ -323,8 +381,20 @@ pub unsafe fn simd_warm_multiply<S : Simd + Copy>(
 		= S::sum_across_n(m0,m1,simd_width,offset_mod_simd);
 	    sum = S::add_elements(sum,part);
 	    m0 = new_m;
-	    x0  = xform.read_next();
-	    i0  = input.read_next();
+	    // x0  = xform.read_next();
+	    // i0  = input.read_next();
+	    x0 = S::read_next(&mut xform_mod_index,
+				 &mut xform_array_index,
+				 xform_array,
+				 xform_size,
+				 &mut xform_ra_size,
+				 &mut xform_ra);
+	    i0 = S::read_next(&mut input_mod_index,
+				 &mut input_array_index,
+				 input_array,
+				 input_size,
+				 &mut input_ra_size,
+				 &mut input_ra);
 	    m1  = S::cross_product(x0,i0); // new m1
 	    dp_counter += simd_width;
 	    // offset_mod_simd unchanged
@@ -332,7 +402,7 @@ pub unsafe fn simd_warm_multiply<S : Simd + Copy>(
 	// above may have set dp_counter to k already.
 	if dp_counter < k {	       // If not, ...
 	    let want = k - dp_counter; // always strictly positive
-	    
+
 	    // eprintln!("Calling sum_across_n with m0 {:?}, m1 {:?}, n {}, offset {}",
 	    //      m0.vec, m1.vec, want, offset_mod_simd);
 	    let (part, new_m) = S::sum_across_n(m0,m1,want,offset_mod_simd);
@@ -343,8 +413,20 @@ pub unsafe fn simd_warm_multiply<S : Simd + Copy>(
 	    if offset_mod_simd + want >= simd_width {
 		// consumed m0 and maybe some of m1 too
 		m0 = new_m;	// nothing left in old m0, so m0 <- m1
-		x0  = xform.read_next();
-		i0  = input.read_next();
+		// x0  = xform.read_next();
+		// i0  = input.read_next();
+		x0 = S::read_next(&mut xform_mod_index,
+				     &mut xform_array_index,
+				     xform_array,
+				     xform_size,
+				     &mut xform_ra_size,
+				     &mut xform_ra);
+		i0 = S::read_next(&mut input_mod_index,
+				     &mut input_array_index,
+				     input_array,
+				     input_size,
+				     &mut input_ra_size,
+				     &mut input_ra);
 		m1  = S::cross_product(x0,i0); // new m1
 	    } else {
 		// got what we needed from m0 but it still has some
@@ -549,8 +631,10 @@ mod tests {
     // }
 
     #[test]
-    #[cfg(any(target_arch = "x86", target_arch = "x86_64"))]
     // test taken from simulator.rs
+    #[cfg(any(target_arch = "x86", target_arch = "x86_64",
+	      all(any(target_arch = "aarch64", target_arch = "arm"),
+		  feature = "arm_vmull")))]
     fn simd_identity_k9_multiply_colwise() {
 	unsafe {
 	    let identity = [
@@ -565,17 +649,17 @@ mod tests {
 		0,0,0, 0,0,0, 0,0,1,
 	    ];
 	    let mut transform =	// mut because of iterator
-		X86SimpleMatrix::<x86::X86u8x16Long0x11b>::new(9,9,true);
+		Matrix::new(9,9,true);
 	    transform.fill(&identity[..]);
 
 	    // 17 is coprime to 9
 	    let mut input =
-		X86SimpleMatrix::<x86::X86u8x16Long0x11b>::new(9,17,false);
+		Matrix::new(9,17,false);
 	    let vec : Vec<u8> = (1u8..=9 * 17).collect();
 	    input.fill(&vec[..]);
-
+	    
 	    let mut output =
-		X86SimpleMatrix::<x86::X86u8x16Long0x11b>::new(9,17,false);
+		Matrix::new(9,17,false);
 
 	    // works if output is stored in colwise format
 	    simd_warm_multiply(&mut transform, &mut input, &mut output);
@@ -585,8 +669,10 @@ mod tests {
     }
 
     #[test]
-    #[cfg(any(target_arch = "x86", target_arch = "x86_64"))]
     // test taken from simulator.rs
+    #[cfg(any(target_arch = "x86", target_arch = "x86_64",
+	      all(any(target_arch = "aarch64", target_arch = "arm"),
+		  feature = "arm_vmull")))]
     fn simd_double_identity() {
 	// seems like lower half of matrix not being output
 	// copy identity matrix down there to test
@@ -612,17 +698,17 @@ mod tests {
 		0,0,0, 0,0,0, 0,0,1,
 	    ];
 	    let mut transform =	// mut because of iterator
-		X86SimpleMatrix::<x86::X86u8x16Long0x11b>::new(18,9,true);
+		Matrix::new(18,9,true);
 	    transform.fill(&double_identity[..]);
 
 	    // 17 is coprime to 9
 	    let mut input =
-		X86SimpleMatrix::<x86::X86u8x16Long0x11b>::new(9,17,false);
+		Matrix::new(9,17,false);
 	    let vec : Vec<u8> = (1u8..=9 * 17).collect();
 	    input.fill(&vec[..]);
 
 	    let mut output =
-		X86SimpleMatrix::<x86::X86u8x16Long0x11b>::new(18,17,true);
+		Matrix::new(18,17,true);
 
 	    // works if output is stored in colwise format
 	    simd_warm_multiply(&mut transform, &mut input, &mut output);
@@ -648,29 +734,31 @@ mod tests {
 
     // test conformance with a variety of matrix sizes
     #[test]
+    #[cfg(any(target_arch = "x86", target_arch = "x86_64",
+	      all(any(target_arch = "aarch64", target_arch = "arm"),
+		  feature = "arm_vmull")))]
     fn test_ref_simd_conformance() {
-
 	let cols = 19;
 	for k in 4..9 {
 	    for n in 4..17 {
 		eprintln!("testing n={}, k={}", n, k);
 		unsafe {
 		    let mut transform =	// mut because of iterator
-			X86SimpleMatrix::<x86::X86u8x16Long0x11b>
+			Matrix
 			::new(n,k,true);
 		    let mut input =
-			X86SimpleMatrix::<x86::X86u8x16Long0x11b>
+			Matrix
 			::new(k,cols,false);
 
 		    transform.fill(&(1u8..).take(n*k).collect::<Vec<u8>>()[..]);
 		    input.fill(&(1u8..).take(k*cols).collect::<Vec<u8>>()[..]);
 
 		    let mut ref_output =
-			X86SimpleMatrix::<x86::X86u8x16Long0x11b>
+			Matrix
 			::new(n,cols,true);
 
 		    let mut simd_output =
-			X86SimpleMatrix::<x86::X86u8x16Long0x11b>
+			Matrix
 			::new(n,cols,true);
 
 		    // do multiply both ways
